import copy, math
from hashlib import new
import enum
from turtle import forward
import torch, dgl
import dgl.function as fn
import torch
from torch import dtype
from torch import embedding
import torch.nn as nn
from .tools import allennlp as util
from tatqa_metric import TaTQAEmAndF1
from .tools.util import FFNLayer
from .tools import allennlp as util
from typing import Dict, List, Tuple
import numpy as np
from reg_hnt.data.file_utils import is_scatter_available
import math
import torch.nn.functional as F
from reg_hnt.reghnt.model_utils import rnn_wrapper, lens2mask, PoolingFunction, FFN
from reg_hnt.reghnt.functions import *
from reg_hnt.reghnt.modeling_tree import TreeNode, Score, TreeAttn, Prediction, TreeEmbedding, TreeBeam, generate_tree_input, GenerateNode, Merge, copy_list
from reg_hnt.reghnt.Vocabulary import *
from reg_hnt.reghnt.masked_cross_entropy import *
from reg_hnt.reghnt.model_functions import *
from reg_hnt.data.pre_order import prefix_eval
from reg_hnt.data.pre_order_ext import pretomid


class RGATLayer(nn.Module):
    def __init__(self, ndim, edim, num_heads=8, feat_drop=0.2):
        super(RGATLayer, self).__init__()
        self.ndim, self.edim = ndim, edim
        self.num_heads = num_heads
        dim = max([ndim, edim])
        self.d_k = dim // self.num_heads
        self.affine_q, self.affine_k, self.affine_v = nn.Linear(self.ndim, dim),\
            nn.Linear(self.ndim, dim, bias=False), nn.Linear(self.ndim, dim, bias=False)
        self.affine_o = nn.Linear(dim, self.ndim)
        self.layernorm = nn.LayerNorm(self.ndim)
        self.feat_dropout = nn.Dropout(p=feat_drop)
        self.ffn = FFN(self.ndim)

    def forward(self, x, lgx, g):
        """ @Params:
                x: node feats, num_nodes x ndim
                lgx: edge feats, num_edges x edim
                g: dgl.graph
        """
        # pre-mapping q/k/v affine
        q, k, v = self.affine_q(self.feat_dropout(x)), self.affine_k(self.feat_dropout(x)), self.affine_v(self.feat_dropout(x))
        e = lgx.view(-1, self.num_heads, self.d_k) if lgx.size(-1) == q.size(-1) else \
            lgx.unsqueeze(1).expand(-1, self.num_heads, -1)
        
        with g.local_scope():
            g.ndata['q'], g.ndata['k'] = q.view(-1, self.num_heads, self.d_k), k.view(-1, self.num_heads, self.d_k)
            g.ndata['v'] = v.view(-1, self.num_heads, self.d_k)
            g.edata['e'] = e
            out_x = self.propagate_attention(g)

        out_x = self.layernorm(x + self.affine_o(out_x.view(-1, self.num_heads * self.d_k)))
        out_x = self.ffn(out_x)
        return out_x, lgx

    def propagate_attention(self, g):
        # Compute attention score
        g.apply_edges(src_sum_edge_mul_dst('k', 'q', 'e', 'score'))
        g.apply_edges(scaled_exp('score', math.sqrt(self.d_k)))
        # Update node state
        g.update_all(src_sum_edge_mul_edge('v', 'e', 'score', 'v'), fn.sum('v', 'wv'))
        g.update_all(fn.copy_edge('score', 'score'), fn.sum('score', 'z'), div_by_z('wv', 'z', 'o'))
        
        out_x = g.ndata['o']
        return out_x



class RGTAT(nn.Module):
    def __init__(self, gnn_hidden_size, gnn_num_layers, num_heads, relation_num):
        super(RGTAT, self).__init__()
        self.num_layers, self.num_heads = gnn_num_layers, num_heads
        self.ndim = gnn_hidden_size
        self.edim = self.ndim // self.num_heads
        self.relation_num = relation_num
        self.relation_embed = nn.Embedding(self.relation_num, self.edim)
        self.dropout = 0.2
        self.gnn_layers = nn.ModuleList([
            RGATLayer(self.ndim, self.edim, num_heads=num_heads, feat_drop=self.dropout)
            for _ in range(self.num_layers)])

    def forward(self, x, batch):
        local_lgx = self.relation_embed(batch['b_relation'])
        local_g = batch['b_graph']
        for i in range(self.num_layers):
            x, local_lgx = self.gnn_layers[i](x, local_lgx, local_g) 
        return x


class SubwordAggregation(nn.Module):
    """ Map subword or wordpieces into one fixed size vector based on aggregation method
    """
    def __init__(self, hidden_size, subword_aggregation='mean-pooling'):
        super(SubwordAggregation, self).__init__()
        self.hidden_size = hidden_size
        self.word_aggregation = PoolingFunction(self.hidden_size, self.hidden_size, method=subword_aggregation)
        self.number_aggregation = PoolingFunction(self.hidden_size, self.hidden_size, method=subword_aggregation)

    def forward(self, inputs, batch):
        bs = inputs.size(0)
        device = inputs.device

        word_mask = batch['word_mask'].bool()
        number_mask = batch['number_mask'].bool()
        word_subword_lens = batch['word_subword_lens']
        number_subword_lens = batch['number_subword_lens']
        word_subword_mask = batch['word_subword_mask']
        number_subword_mask = batch['number_subword_mask']

        old_word, old_number = inputs.masked_select(word_mask.unsqueeze(-1)), inputs.masked_select(number_mask.unsqueeze(-1))
        words = old_word.new_zeros(word_subword_lens.size(0), max(batch['word_subword_lens']).item(), self.hidden_size)
        try:
            words = words.masked_scatter_(word_subword_mask.unsqueeze(-1), old_word)
        except:
            import pdb; pdb.set_trace()
        numbers = old_number.new_zeros(number_subword_lens.size(0), max(batch['number_subword_lens']).item(), self.hidden_size)
        try:
            numbers = numbers.masked_scatter_(number_subword_mask.unsqueeze(-1), old_number)
        except:
            import pdb; pdb.set_trace()

        words = self.word_aggregation(words, mask=word_subword_mask)
        numbers = self.word_aggregation(numbers, mask=number_subword_mask)
        
        b_word_word_mask = batch['b_word_word_mask']
        b_number_word_mask = batch['b_number_word_mask']
        new_inputs = torch.zeros([b_word_word_mask.size(0), self.hidden_size]).to(device)
        new_inputs = new_inputs.masked_scatter_(b_word_word_mask.unsqueeze(-1), words)
        new_inputs = new_inputs.masked_scatter_(b_number_word_mask.unsqueeze(-1), numbers)
        
        
        inputs = torch.zeros([bs, batch['pl_mask']['b'].size(1), self.hidden_size]).to(device)
        inputs = inputs.masked_scatter_(batch['pl_mask']['b'].unsqueeze(-1), new_inputs)
        
        old_questions, old_tables, old_paragraphs = inputs.masked_select(batch['pl_mask']['q'].unsqueeze(-1)), \
            inputs.masked_select(batch['pl_mask']['t'].unsqueeze(-1)), inputs.masked_select(batch['pl_mask']['p'].unsqueeze(-1))
        questions = old_questions.new_zeros(bs, batch['max_len']['question'], self.hidden_size)
        questions = questions.masked_scatter_(batch['pl_mask']['question'].unsqueeze(-1), old_questions)
        tables = old_tables.new_zeros(bs, batch['max_len']['table_word'], self.hidden_size)
        tables = tables.masked_scatter_(batch['pl_mask']['table'].unsqueeze(-1), old_tables)
        paragraphs = old_paragraphs.new_zeros(bs, batch['max_len']['paragraph'], self.hidden_size)
        paragraphs = paragraphs.masked_scatter_(batch['pl_mask']['paragraph'].unsqueeze(-1), old_paragraphs)
        
        return questions, tables, paragraphs


class InputRNNLayer(nn.Module):
    def __init__(self, input_size, hidden_size, cell='lstm', schema_aggregation='attentive-pooling', share_lstm=False):
        super(InputRNNLayer, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.cell = cell.upper()
        self.question_lstm = getattr(nn, self.cell)(self.input_size, self.hidden_size // 2, num_layers=1, bidirectional=True, batch_first=True)
        self.table_lstm = getattr(nn, self.cell)(self.input_size, self.hidden_size // 2, num_layers=1, bidirectional=True, batch_first=True)
        self.paragraph_lstm = getattr(nn, self.cell)(self.input_size, self.hidden_size // 2, num_layers=1, bidirectional=True, batch_first=True)
        self.schema_aggregation = schema_aggregation
        if self.schema_aggregation != 'head+tail':
            self.aggregation = PoolingFunction(self.hidden_size, self.hidden_size, method=schema_aggregation)

    def forward(self, input_dict, batch):
        """
            for question sentence, forward into a bidirectional LSTM to get contextual info and sequential dependence
            for schema phrase, extract representation for each phrase by concatenating head+tail vectors,
            batch.question_lens, batch.table_word_lens, batch.column_word_lens are used
        """

        device = input_dict['question'].device

        questions, _ = rnn_wrapper(self.question_lstm, input_dict['question'], batch['lens']['question'], cell=self.cell)
        questions = questions.contiguous().view(-1, self.hidden_size)[lens2mask(batch['lens']['question']).contiguous().view(-1)]

        paragraphs, _ = rnn_wrapper(self.paragraph_lstm, input_dict['paragraph'], batch['lens']['paragraph'], cell=self.cell)
        paragraphs = paragraphs.contiguous().view(-1, self.hidden_size)[lens2mask(batch['lens']['paragraph']).contiguous().view(-1)]
        
        
        old_tables, _ = rnn_wrapper(self.table_lstm, input_dict['table'], batch['lens']['table_word'], cell=self.cell)
        old_tables = old_tables.contiguous().view(-1, self.hidden_size)[lens2mask(batch['lens']['table_word']).contiguous().view(-1)]
        t_w_t_n_mask = batch['t_w_t_n_mask']
        
        new_tables = torch.zeros([t_w_t_n_mask.size(0), t_w_t_n_mask.size(1), self.hidden_size]).to(device)
        new_tables = new_tables.masked_scatter_(t_w_t_n_mask.unsqueeze(-1), old_tables)
        tables = self.aggregation(new_tables, mask=t_w_t_n_mask)

        
        questions = questions.split(batch['lens']['question'].tolist(), dim=0)
        tables = tables.split(batch['lens']['table'].tolist(), dim=0)
        paragraphs = paragraphs.split(batch['lens']['paragraph'].tolist(), dim=0)

        outputs = [th for q_t_c in zip(questions, tables, paragraphs) for th in q_t_c]
        outputs = torch.cat(outputs, dim=0)

        return outputs



def generate_question_3split(max_seq,max_len,split=3):
    punc_list=[",","：","；","？","！","，","“","”",",",".","?","，","。","？","．","；","｡"]
    question_mask = []
    split=2
    for i in range(max_seq):
        question_mask.append([0 for _ in range(split)])
    for i in range(max_seq,max_len):
        question_mask.append([1 for _ in range(split)])
    
    split_size1 = int(max_seq * 0.5)

    for i in range(0,split_size1):
        question_mask[i][1]=1

    for i in range(split_size1,max_seq):
        question_mask[i][0]=1
    return question_mask



    
class Arithmetic(nn.Module):
    def __init__(self, hidden_size):
        super(Arithmetic, self).__init__()
        self.hidden_size = hidden_size
        embedding_size = hidden_size
        self.embedding_size = embedding_size
        self.prediction = Prediction(hidden_size, op_nums=len(OPERATOR_LIST),input_size=len(CONST_LIST2))
        self.generate = GenerateNode(hidden_size, op_nums=len(OPERATOR_LIST), embedding_size=embedding_size)
        self.merge = Merge(hidden_size=hidden_size, embedding_size=embedding_size)
        
        
    def get_all_number_encoder_outputs(self, encoder_outputs, num_pos, batch_size, num_size, hidden_size):
        indices = list()
        sen_len = encoder_outputs.size(0)
        masked_index = []
        temp_1 = [1 for _ in range(hidden_size)]
        temp_0 = [0 for _ in range(hidden_size)]
        for b in range(batch_size):
            for i in num_pos[b]:
                indices.append(i + b * sen_len)
                masked_index.append(temp_0)
            indices += [0 for _ in range(len(num_pos[b]), num_size)]
            masked_index += [temp_1 for _ in range(len(num_pos[b]), num_size)]
        indices = torch.LongTensor(indices).cuda()
        masked_index = torch.ByteTensor(masked_index).cuda()
        masked_index = masked_index.view(batch_size, num_size, hidden_size)
        all_outputs = encoder_outputs.transpose(0, 1).contiguous()
        all_embedding = all_outputs.view(-1, encoder_outputs.size(2))
        all_num = all_embedding.index_select(0, indices)
        all_num = all_num.view(batch_size, num_size, hidden_size)
        masked_index=masked_index.bool()
        
        return all_num.masked_fill_(masked_index, 0.0),indices,masked_index
    
    def forward(self, sequence_output, hidden, batch, arithmetic_labels):
        target_batch = [output for i, output in enumerate(batch['outputs']) if arithmetic_labels[i]==1]

        cls_output = sequence_output[:, 0, :]
        batch_size = cls_output.size(0)

        target_length = [len(item) for item in target_batch]
        max_target_length = max([len(item) for item in target_batch])
        
        all_node_outputs = []
        
        question_lens = batch['lens']['question'].masked_select(torch.tensor(arithmetic_labels).bool())
        table_lens = batch['lens']['table'].masked_select(torch.tensor(arithmetic_labels).bool())
        paragraph_lens = batch['lens']['paragraph'].masked_select(torch.tensor(arithmetic_labels).bool())
        b_lens = torch.zeros([batch_size]).cuda()

        table_number_index = [output for i, output in enumerate(batch['table_number_index']) if arithmetic_labels[i]==1]
        paragraph_number_index = [output for i, output in enumerate(batch['paragraph_number_index']) if arithmetic_labels[i]==1]

        num_pos = []
        for i in range(batch_size):
            numbers = []
            numbers += [batch_number.item() + question_lens[i].item() for batch_number in table_number_index[i]]
            numbers += [batch_number.item() +question_lens[i].item()+table_lens[i].item() for batch_number in paragraph_number_index[i]]
            num_pos.append(numbers)

        copy_num_len = [len(_) for _ in num_pos]
        max_num_size = max(copy_num_len)
        
        for i in range(batch_size):
            b_lens[i] = question_lens[i] + table_lens[i] + paragraph_lens[i]
        
        b_lens = list(map(int, b_lens.tolist()))
        max_len = max(b_lens)
        
        change_size = False
        if max_len != hidden.size(1):
            change_size = True
            
            
        if change_size:
            hidden = hidden[:, :max_len, :]
        encoder_outputs = hidden.transpose(0,1).contiguous()
        

        all_nums_encoder_outputs, indices, masked_index = self.get_all_number_encoder_outputs(encoder_outputs, num_pos, batch_size, max_num_size,self.hidden_size)
        
        
        padding_hidden = torch.zeros([1, self.hidden_size]).cuda()
        
        b_mask = batch['mask']['b'].masked_select(torch.tensor(arithmetic_labels).unsqueeze(-1).bool()).view(batch_size, -1)
        if change_size:
            b_mask = b_mask[:, :max_len]
        
        seq_mask = (b_mask != False).to(dtype=torch.uint8)
        
        num_mask = []
        max_num_size = max_num_size + len(CONST_LIST2)
        for i in copy_num_len:
            d = i + len(CONST_LIST2)
            num_mask.append([0] * d + [1] * (max_num_size - d))
        num_mask = torch.ByteTensor(num_mask).cuda()
        
        
        question_mask = []

        for b in range(batch_size):
            ques_mask_this = generate_question_3split(b_lens[b], max_len)
            question_mask.append(ques_mask_this)
        question_mask = torch.ByteTensor(question_mask).cuda()

        
        nums_stack_batch = [[] for _ in range(batch_size)]
        
        for target_batch_i in target_batch:
            target_batch_i.extend([0 for _ in range(max_target_length-len(target_batch_i))])

        target = torch.LongTensor(target_batch).cuda().transpose(0, 1)

        node_stacks = [[TreeNode(_)] for _ in cls_output.split(1, dim=0)]
        num_start = len(OPERATOR_LIST)
        
        embeddings_stacks = [[] for _ in range(batch_size)]
        left_childs = [None for _ in range(batch_size)]

        for t in range(max_target_length):

            num_score, op, current_embeddings, current_context, current_nums_embeddings,current_attn = self.prediction(
                node_stacks, left_childs, encoder_outputs, all_nums_encoder_outputs, padding_hidden, seq_mask, num_mask,indices,masked_index,question_mask)

            outputs = torch.cat((op, num_score), 1)
            all_node_outputs.append(outputs)

            target_t, generate_input = generate_tree_input(target[t].tolist(), outputs, nums_stack_batch, num_start)
            target[t] = target_t
            left_child, right_child, node_label = self.generate(current_embeddings, generate_input, current_context)
            left_childs = []
            for idx, l, r, node_stack, i, o in zip(range(batch_size), left_child.split(1), right_child.split(1),
                                               node_stacks, target[t].tolist(), embeddings_stacks):
                if len(node_stack) != 0:
                    node = node_stack.pop()
                else:
                    left_childs.append(None)
                    continue
                # 是运算符
                if i < num_start:
                    node_stack.append(TreeNode(r))
                    node_stack.append(TreeNode(l, left_flag=True))
                    o.append(TreeEmbedding(node_label[idx].unsqueeze(0), False))
                # 是数字
                else:
                    current_num = current_nums_embeddings[idx, i- num_start].unsqueeze(0)
                    while len(o) > 0 and o[-1].terminal:
                        sub_stree = o.pop()
                        op = o.pop()
                        current_num = self.merge(op.embedding, sub_stree.embedding, current_num)
                    o.append(TreeEmbedding(current_num, True))
                if len(o) > 0 and o[-1].terminal:
                    left_childs.append(o[-1].embedding)
                else:
                    left_childs.append(None)
        
        try:
            all_node_outputs = torch.stack(all_node_outputs, dim=1)
        except:
            import pdb; pdb.set_trace()

        target = target.transpose(0, 1).contiguous()

        loss = masked_cross_entropy(all_node_outputs, target, target_length)
        return loss, all_node_outputs, target

    def predict(self, sequence_output, hidden, batch, arithmetic_labels, predicted_scale_class, gold_answer):
        MAX_LENGTH = 28
        beam_size = 3
        

        cls_output = sequence_output[:, 0, :]
        batch_size = cls_output.size(0)
        
        question_lens = batch['lens']['question'].masked_select(torch.tensor(arithmetic_labels).bool())
        table_lens = batch['lens']['table'].masked_select(torch.tensor(arithmetic_labels).bool())
        paragraph_lens = batch['lens']['paragraph'].masked_select(torch.tensor(arithmetic_labels).bool())
        b_lens = torch.zeros([batch_size]).cuda()

        table_number_index = [output for i, output in enumerate(batch['table_number_index']) if arithmetic_labels[i]==1]
        paragraph_number_index = [output for i, output in enumerate(batch['paragraph_number_index']) if arithmetic_labels[i]==1]
        b_tokens = [output for i, output in enumerate(batch['b_tokens']) if arithmetic_labels[i]==1]
        b_question_lens = [output for i, output in enumerate(batch['lens']['question']) if arithmetic_labels[i]==1]
        num_pos = []
        for i in range(batch_size):
            numbers = []
            numbers += [batch_number.item() + question_lens[i].item() for batch_number in table_number_index[i]]
            numbers += [batch_number.item() +question_lens[i].item()+table_lens[i].item() for batch_number in paragraph_number_index[i]]
            num_pos.append(numbers)
        
        copy_num_len = [len(_) for _ in num_pos]
        max_num_size = max(copy_num_len)
        
        for i in range(batch_size):
            b_lens[i] = question_lens[i] + table_lens[i] + paragraph_lens[i]
        
        b_lens = list(map(int, b_lens.tolist()))
        max_len = max(b_lens)
        
        change_size = False
        if max_len != hidden.size(1):
            change_size = True
            
            
        if change_size:
            hidden = hidden[:, :max_len, :]
        encoder_outputs = hidden.transpose(0,1).contiguous()
        

        all_nums_encoder_outputs, indices, masked_index = self.get_all_number_encoder_outputs(encoder_outputs, num_pos, batch_size, max_num_size,self.hidden_size)
        
        
        padding_hidden = torch.zeros([1, self.hidden_size]).cuda()
        
        b_mask = batch['mask']['b'].masked_select(torch.tensor(arithmetic_labels).unsqueeze(-1).bool()).view(batch_size, -1)
        if change_size:
            b_mask = b_mask[:, :max_len]
        
        seq_mask = (b_mask != False).to(dtype=torch.uint8)
        
        num_mask = []
        max_num_size = max_num_size + len(CONST_LIST2)
        for i in copy_num_len:
            d = i + len(CONST_LIST2)
            num_mask.append([0] * d + [1] * (max_num_size - d))
        num_mask = torch.ByteTensor(num_mask).cuda()
        
        
        question_mask = []

        for b in range(batch_size):
            ques_mask_this = generate_question_3split(b_lens[b], max_len)
            question_mask.append(ques_mask_this)
        question_mask = torch.ByteTensor(question_mask).cuda()



        node_stacks = [[TreeNode(_)] for _ in cls_output.split(1, dim=0)]
        num_start = len(OPERATOR_LIST)
        
        embeddings_stacks = [[] for _ in range(batch_size)]
        left_childs = [None for _ in range(batch_size)]
        
        beams = [TreeBeam(0.0, node_stacks, embeddings_stacks, left_childs, [])]
        
        for t in range(MAX_LENGTH):
            current_beams = []

            while len(beams) > 0:
                b = beams.pop()
                if len(b.node_stack[0]) == 0:
                    current_beams.append(b)
                    continue
                
                left_childs = b.left_childs
                            # num_score [B, 2 + num_num]  op [B, op_num]  current_embeddings [B,1,H]  current_context [B,1,H] current_nums_embeddings [B, 2+num_num,H]
                num_score, op, current_embeddings, current_context, current_nums_embeddings,current_attn = self.prediction(
                    b.node_stack, left_childs, encoder_outputs, all_nums_encoder_outputs, padding_hidden,
                    seq_mask, num_mask,indices, masked_index, question_mask)
                out_score = nn.functional.log_softmax(torch.cat((op, num_score), dim=1), dim=1)
                # topv 分数    topi 下标
                topv, topi = out_score.topk(beam_size)
                
                
                for tv, ti in zip(topv.split(1, dim=1), topi.split(1, dim=1)):
                    current_node_stack = copy_list(b.node_stack)
                    current_left_childs = []
                    current_embeddings_stacks = copy_list(b.embedding_stack)
                    current_out = copy.deepcopy(b.out)

                    out_token = int(ti)
                    current_out.append(out_token)

                    node = current_node_stack[0].pop()

                    if out_token < num_start:
                        generate_input = torch.LongTensor([out_token]).cuda()
                        left_child, right_child, node_label = self.generate(current_embeddings, generate_input, current_context)

                        current_node_stack[0].append(TreeNode(right_child))
                        current_node_stack[0].append(TreeNode(left_child, left_flag=True))

                        current_embeddings_stacks[0].append(TreeEmbedding(node_label[0].unsqueeze(0), False))
                    else:
                        current_num = current_nums_embeddings[0, out_token - num_start].unsqueeze(0)

                        while len(current_embeddings_stacks[0]) > 0 and current_embeddings_stacks[0][-1].terminal:
                            sub_stree = current_embeddings_stacks[0].pop()
                            op = current_embeddings_stacks[0].pop()
                            current_num = self.merge(op.embedding, sub_stree.embedding, current_num)
                        current_embeddings_stacks[0].append(TreeEmbedding(current_num, True))
                    if len(current_embeddings_stacks[0]) > 0 and current_embeddings_stacks[0][-1].terminal:
                        current_left_childs.append(current_embeddings_stacks[0][-1].embedding)
                    else:
                        current_left_childs.append(None)
                    current_beams.append(TreeBeam(b.score+float(tv), current_node_stack, current_embeddings_stacks,
                                                current_left_childs, current_out))
            beams = sorted(current_beams, key=lambda x: x.score, reverse=True)
            beams = beams[:beam_size]
            flag = True
            for b in beams:
                if len(b.node_stack[0]) != 0:
                    flag = False
            if flag:
                break
        del seq_mask,padding_hidden,num_mask
        torch.cuda.empty_cache()
        predict_answer_id = beams[0].out
        numbers = OPERATOR_LIST + CONST_LIST2 + batch['question_number'][arithmetic_labels.index(1)]  + batch['table_number'][arithmetic_labels.index(1)] + batch['paragraph_number'][arithmetic_labels.index(1)]
        question_id = batch['question_id'][arithmetic_labels.index(1)]
        
        numbers_scale = [-1 for _ in range(len(OPERATOR_LIST)+len(CONST_LIST2))] + batch['t_scale'][0].tolist() + batch['p_scale'][0].tolist()
        zero_count = predict_answer_id.count(0)
        
        if len(predict_answer_id) == 0 or predict_answer_id is None or zero_count == len(predict_answer_id):
            answer = -1
            return answer
        else:
            while predict_answer_id[-1] == 0:
                predict_answer_id.pop(-1)
            if len(predict_answer_id) == 1 and predict_answer_id[0] >= len(OPERATOR_LIST) + len(CONST_LIST2)+ len(batch['question_number'][arithmetic_labels.index(1)]) and predict_answer_id[0] < len(OPERATOR_LIST) + len(CONST_LIST2) + len(batch['question_number'][arithmetic_labels.index(1)]) + len(batch['table_number'][arithmetic_labels.index(1)]):
                answer_id = predict_answer_id[0]-len(OPERATOR_LIST) - len(CONST_LIST2) - len(batch['question_number'][arithmetic_labels.index(1)])
                table_token_id = table_number_index[0].tolist()[answer_id]
                answer = b_tokens[0][question_lens[0].item() + table_token_id]
                return answer
            else:
                pre_answer = list(map(lambda x: numbers[x], predict_answer_id))
                pre_answer = list(map(lambda x: str(x) if not isinstance(x, str) else x, pre_answer))

                prefix_expr = ' '.join(pre_answer)
                
                try:
                    answer = prefix_eval(prefix_expr)
                except:
                    answer = 0
            
                
                if predicted_scale_class == 4:
                    answer = np.around(answer*100.0, 2)
                else:
                    answer = np.around(answer, 2)
                # answer, gold_answer['answer'], batch['question_id'], predict_answer_id, predict_answer_id, prefix_expr
                # predict_answer_id, prefix_expr
                # try:
                #     if gold_answer['answer'] + answer == 0:
                #         import pdb; pdb.set_trace()
                # except:
                #     answer = answer
                return answer
                        


def scale_judge(predict_answer_id, numbers, numbers_scale, batch):
    predicted_scale_class = 0
    question = batch['questions'][0]
    if 'percentage' in question:
        predicted_scale_class = 4
        return predicted_scale_class
    predict_number = [id for id in predict_answer_id if id >= len(OPERATOR_LIST)+len(CONST_LIST2)]
    predict_number_scale = list(map(lambda x: numbers_scale[x], predict_number))

    predict_number_scale_set = list(set(predict_number_scale))
    if len(predict_number_scale_set) == 0:
        return predicted_scale_class
    if len(predict_number_scale_set) == 2:
        shuzhi_big = max(predict_number_scale_set)
        maxscale = max(predict_number_scale,key=predict_number_scale.count)
        if maxscale != shuzhi_big:
            predicted_scale_class = maxscale
        else:
            predicted_scale_class = shuzhi_big
        return predicted_scale_class
    if len(predict_number_scale_set) == 1:
        return predict_number_scale_set[0]
    
    return -1

class OtherModel(nn.Module):
    def __init__(self, hidden_size):
        super(OtherModel, self).__init__()
        self.hidden_size = hidden_size
        self.tag_predictor = FFNLayer(hidden_size, hidden_size, 2, 0.2)
        self.NLLLoss = nn.NLLLoss(reduction="sum")

    def forward(self, sequence_output, hidden, batch, other_labels):
        batch_size = hidden.size(0)
        question_lens = batch['lens']['question'].masked_select(torch.tensor(other_labels).bool())
        table_lens = batch['lens']['table'].masked_select(torch.tensor(other_labels).bool())
        paragraph_lens = batch['lens']['paragraph'].masked_select(torch.tensor(other_labels).bool())
        
        b_lens = torch.zeros([batch_size]).cuda()
        for i in range(batch_size):
            b_lens[i] = question_lens[i] + table_lens[i] + paragraph_lens[i]
        
        b_lens = list(map(int, b_lens.tolist()))
        max_len = max(b_lens)
        
        change_size = False
        if max_len != hidden.size(1):
            change_size = True
        
        if change_size:
            hidden = hidden[:, :max_len, :]
        
        b_mask = batch['mask']['b'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1)
        question_mask = batch['mask']['question'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).long()
        table_mask = batch['mask']['table'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).long()
        paragraph_mask = batch['mask']['paragraph'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).long()
        tags = batch['b_tags'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).float()
        
        if change_size:
            b_mask = b_mask[:, :max_len]
            question_mask = question_mask[:, :max_len]
            table_mask = table_mask[:, :max_len]
            paragraph_mask = paragraph_mask[:, :max_len]
            tags = tags[:, :max_len]


        table_sequence_output = util.replace_masked_values(hidden, table_mask.unsqueeze(-1), 0)
        table_tag_prediction = self.tag_predictor(table_sequence_output)
        table_tag_prediction = util.masked_log_softmax(table_tag_prediction, mask=None)
        table_tag_prediction = util.replace_masked_values(table_tag_prediction, table_mask.unsqueeze(-1), 0)
        table_tag_labels = util.replace_masked_values(tags, table_mask, 0)
        
        paragraph_sequence_output = util.replace_masked_values(hidden, paragraph_mask.unsqueeze(-1), 0)
        paragraph_tag_prediction = self.tag_predictor(paragraph_sequence_output)
        paragraph_tag_prediction = util.masked_log_softmax(paragraph_tag_prediction, mask=None)
        paragraph_tag_prediction = util.replace_masked_values(paragraph_tag_prediction, paragraph_mask.unsqueeze(-1), 0)
        paragraph_tag_labels = util.replace_masked_values(tags, paragraph_mask, 0)


        table_tag_prediction = table_tag_prediction.transpose(1, 2)
        table_tag_prediction_loss = self.NLLLoss(table_tag_prediction, table_tag_labels.long())
        
        paragraph_tag_prediction = paragraph_tag_prediction.transpose(1, 2)
        paragraph_token_tag_prediction_loss = self.NLLLoss(paragraph_tag_prediction, paragraph_tag_labels.long())
        
        other_loss = table_tag_prediction_loss + paragraph_token_tag_prediction_loss
        
        return other_loss, table_tag_prediction, paragraph_tag_prediction


    def forward2(self, sequence_output, hidden, batch, other_labels):
        batch_size = hidden.size(0)
        question_lens = batch['lens']['question'].masked_select(torch.tensor(other_labels).bool())
        table_lens = batch['lens']['table'].masked_select(torch.tensor(other_labels).bool())
        paragraph_lens = batch['lens']['paragraph'].masked_select(torch.tensor(other_labels).bool())
        
        b_lens = torch.zeros([batch_size]).cuda()
        for i in range(batch_size):
            b_lens[i] = question_lens[i] + table_lens[i] + paragraph_lens[i]
        
        b_lens = list(map(int, b_lens.tolist()))
        max_len = max(b_lens)
        
        change_size = False
        if max_len != hidden.size(1):
            change_size = True
        
        if change_size:
            hidden = hidden[:, :max_len, :]
        
        b_mask = batch['mask']['b'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1)
        question_mask = batch['mask']['question'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).long()
        table_mask = batch['mask']['table'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).long()
        paragraph_mask = batch['mask']['paragraph'].masked_select(torch.tensor(other_labels).unsqueeze(-1).bool()).view(batch_size, -1).long()
        
        if change_size:
            b_mask = b_mask[:, :max_len]
            question_mask = question_mask[:, :max_len]
            table_mask = table_mask[:, :max_len]
            paragraph_mask = paragraph_mask[:, :max_len]

        table_sequence_output = util.replace_masked_values(hidden, table_mask.unsqueeze(-1), 0)
        table_tag_prediction = self.tag_predictor(table_sequence_output)
        table_tag_prediction = util.masked_log_softmax(table_tag_prediction, mask=None)
        table_tag_prediction = util.replace_masked_values(table_tag_prediction, table_mask.unsqueeze(-1), 0)

        
        paragraph_sequence_output = util.replace_masked_values(hidden, paragraph_mask.unsqueeze(-1), 0)
        paragraph_tag_prediction = self.tag_predictor(paragraph_sequence_output)
        paragraph_tag_prediction = util.masked_log_softmax(paragraph_tag_prediction, mask=None)
        paragraph_tag_prediction = util.replace_masked_values(paragraph_tag_prediction, paragraph_mask.unsqueeze(-1), 0)

    
        table_tag_prediction = table_tag_prediction.transpose(1, 2)
        
        paragraph_tag_prediction = paragraph_tag_prediction.transpose(1, 2)
        
        other_loss = 0
        
        return other_loss, table_tag_prediction, paragraph_tag_prediction



    def predict(self, hidden, batch, other_labels, operator_class, table_tag_prediction, paragraph_tag_prediction):

        table_tag_prediction = table_tag_prediction.transpose(1, 2)
        table_tag_prediction_score = table_tag_prediction[:, :, 1]
        table_tag_prediction = torch.argmax(table_tag_prediction, dim=-1).float()
        table_tag_prediction = table_tag_prediction.detach().cpu().numpy()
        table_tag_prediction_score = table_tag_prediction_score.detach().cpu().numpy()
        
        paragraph_tag_prediction = paragraph_tag_prediction.transpose(1, 2)
        paragraph_tag_prediction_score = paragraph_tag_prediction[:, :, 1]
        paragraph_tag_prediction = torch.argmax(paragraph_tag_prediction, dim=-1).float() 
        paragraph_tag_prediction = paragraph_tag_prediction.detach().cpu().numpy()
        paragraph_tag_prediction_score = paragraph_tag_prediction_score.detach().cpu().numpy()

        global_tokens = batch['b_tokens'][other_labels.index(1)]
        answer = None
        if operator_class == OPERATOR_CLASSES["SPAN"]:
            answer = get_single_span_tokens_from_table_and_paragraph(
                table_tag_prediction,
                table_tag_prediction_score,
                paragraph_tag_prediction,
                paragraph_tag_prediction_score,
                global_tokens
            )
            answer = sorted(answer)
        elif operator_class == OPERATOR_CLASSES['MULTI_SPAN']:
            paragraph_selected_span_tokens = \
                get_span_tokens_from_paragraph(paragraph_tag_prediction, global_tokens)
            table_selected_tokens = \
                get_span_tokens_from_table(table_tag_prediction, global_tokens)
            answer =  paragraph_selected_span_tokens + table_selected_tokens
            answer = sorted(answer)
        elif operator_class == OPERATOR_CLASSES['COUNT']:
            paragraph_selected_tokens = \
                get_span_tokens_from_paragraph(paragraph_tag_prediction, global_tokens)
            table_selected_tokens = \
                get_span_tokens_from_table(table_tag_prediction, global_tokens)
            answer = len(paragraph_selected_tokens) + len(table_selected_tokens)
        
        if answer is None or answer == []:
            answer = -1
        if answer == 0 and operator_class == OPERATOR_CLASSES['COUNT']:
            answer = 1
        return answer







import json
import os
from typing import Optional, List, Tuple, Dict

import peft
import torch
from peft import PeftModel, LoraConfig, TaskType, get_peft_model
from transformers.trainer import WEIGHTS_NAME, WEIGHTS_INDEX_NAME, TRAINER_STATE_NAME
from transformers import PreTrainedModel, PreTrainedTokenizer, AutoTokenizer, AutoConfig, BitsAndBytesConfig, \
    AutoModelForCausalLM
from transformers.modeling_utils import load_sharded_checkpoint
from transformers.utils.versions import require_version

from .utils import get_logger, ModelArguments, FinetuningArguments
from .utils import VALUE_HEAD_FILE_NAME


logger = get_logger(__name__)


def prepare_model_for_training(
        model: PreTrainedModel,
        finetuning_type: str,
        output_embedding_layer_name: Optional[str] = "lm_head",
        use_gradient_checkpointing: Optional[bool] = True,
        layer_norm_names: Optional[List[str]] = ["norm", "ln_f"] # for LLaMA and BLOOM setting
) -> PreTrainedModel:

    for name, param in model.named_parameters():
        if param.ndim == 1 and any(layer_norm_name in name for layer_norm_name in layer_norm_names):
            param.data = param.data.to(torch.float32)

    if use_gradient_checkpointing:
        if hasattr(model, "enable_input_require_grads"):
            model.enable_input_require_grads()
        else:
            def make_inputs_require_grad(module, input, output):
                output.requires_grad_(True)
            model.get_input_embeddings().register_forward_hook(make_inputs_require_grad)

        model.gradient_checkpointing_enable()
        model.config.use_cache = False # turn off when gradient checkpointing is enabled

    if finetuning_type != "full" and hasattr(model, output_embedding_layer_name):
        output_embedding_layer: torch.nn.Linear = getattr(model, output_embedding_layer_name)
        input_dtype = output_embedding_layer.weight.dtype

        class CastOutputToFloat(torch.nn.Sequential):

            def forward(self, x: torch.Tensor) -> torch.Tensor:
                return super().forward(x.to(input_dtype)).to(torch.float32)

        setattr(model, output_embedding_layer_name, CastOutputToFloat(output_embedding_layer))

    return model


def print_trainable_params(model: torch.nn.Module) -> None:
    trainable_params, all_param = 0, 0
    for param in model.parameters():
        num_params = param.numel()
        # if using DS Zero 3 and the weights are initialized empty
        if num_params == 0 and hasattr(param, "ds_numel"):
            num_params = param.ds_numel
        all_param += num_params
        if param.requires_grad:
            trainable_params += num_params
    print("trainable params: {:d} || all params: {:d} || trainable%: {:.4f}".format(
                trainable_params, all_param, 100 * trainable_params / all_param))


def get_state_dict(model: torch.nn.Module) -> Dict[str, torch.Tensor]: # get state dict containing trainable parameters
    state_dict = model.state_dict()
    filtered_state_dict = {}

    for k, v in model.named_parameters():
        if v.requires_grad:
            filtered_state_dict[k] = state_dict[k].cpu().clone().detach()

    return filtered_state_dict


def load_trainable_params(model: torch.nn.Module, checkpoint_dir: os.PathLike) -> bool:
    weights_file = os.path.join(checkpoint_dir, WEIGHTS_NAME)
    if os.path.exists(weights_file):
        model_state_dict = torch.load(weights_file, map_location="cpu")
        model.load_state_dict(model_state_dict, strict=False) # skip missing keys
    elif os.path.exists(os.path.join(checkpoint_dir, WEIGHTS_INDEX_NAME)):
        load_sharded_checkpoint(model, checkpoint_dir, strict=False)
    else:
        logger.warning("Provided path ({}) does not contain pre-trained weights.".format(checkpoint_dir))
        return False
    return True


def load_valuehead_params(model: torch.nn.Module, checkpoint_dir: os.PathLike) -> bool:
    valuehead_file = os.path.join(checkpoint_dir, VALUE_HEAD_FILE_NAME)
    if not os.path.exists(valuehead_file):
        logger.warning("Provided path ({}) does not contain valuehead weights.".format(checkpoint_dir))
        return False
    valuehead_state_dict = torch.load(valuehead_file, map_location="cpu")
    model.register_buffer("reward_head_weight", valuehead_state_dict["summary.weight"])
    model.register_buffer("reward_head_bias", valuehead_state_dict["summary.bias"])
    model.register_buffer("default_head_weight", torch.zeros_like(valuehead_state_dict["summary.weight"]))
    model.register_buffer("default_head_bias", torch.zeros_like(valuehead_state_dict["summary.bias"]))
    return True


def smooth(scalars: List[float], weight: Optional[float] = 0.9) -> List[float]:
    r"""
    EMA implementation according to TensorBoard.
    """
    last = scalars[0]
    smoothed = list()
    for next_val in scalars:
        smoothed_val = last * weight + (1 - weight) * next_val
        smoothed.append(smoothed_val)
        last = smoothed_val
    return smoothed


def _init_adapter(
        model: PreTrainedModel,
        model_args: ModelArguments,
        finetuning_args: FinetuningArguments,
        is_trainable: bool,
        is_mergeable: bool
) -> PreTrainedModel:
    r"""
    Initializes the adapters.

    Support full-parameter, freeze and LoRA training.

    Note that the trainable parameters must be cast to float32.
    """

    if finetuning_args.finetuning_type == "none" and is_trainable:
        raise ValueError("You cannot use finetuning_type=none while training.")

    if finetuning_args.finetuning_type == "full":
        logger.info("Fine-tuning method: Full")
        model = model.float()

    if finetuning_args.finetuning_type == "freeze":
        logger.info("Fine-tuning method: Freeze")

        for name, param in model.named_parameters():
            if not any(trainable_layer in name for trainable_layer in finetuning_args.trainable_layers):
                param.requires_grad_(False)
            else:
                param.data = param.data.to(torch.float32)

        if model_args.checkpoint_dir is not None:
            assert load_trainable_params(model, model_args.checkpoint_dir[0]), "Model checkpoint is not correctly loaded."

    if finetuning_args.finetuning_type == "lora":
        logger.info("Fine-tuning method: LoRA")
        lastest_checkpoint = None

        if model_args.checkpoint_dir is not None:
            assert os.path.exists(os.path.join(model_args.checkpoint_dir[0], peft.utils.WEIGHTS_NAME)), \
                "Provided path ({}) does not contain a LoRA weight.".format(model_args.checkpoint_dir[0])
            assert os.path.exists(os.path.join(model_args.checkpoint_dir[0], peft.utils.CONFIG_NAME)), \
                "The given checkpoint may be not a LoRA checkpoint, please specify `--finetuning_type full/freeze` instead."

            if (is_trainable and model_args.resume_lora_training) or (not is_mergeable): # continually train on the lora weights
                checkpoints_to_merge, lastest_checkpoint = model_args.checkpoint_dir[:-1], model_args.checkpoint_dir[-1]
            else:
                checkpoints_to_merge = model_args.checkpoint_dir

            for checkpoint in checkpoints_to_merge:
                model = PeftModel.from_pretrained(model, checkpoint)
                model = model.merge_and_unload()

            if len(checkpoints_to_merge) > 0:
                logger.info("Merged {} model checkpoint(s).".format(len(checkpoints_to_merge)))

            if lastest_checkpoint is not None: # resume lora training or quantized inference
                model = PeftModel.from_pretrained(model, lastest_checkpoint, is_trainable=is_trainable)

        if is_trainable and lastest_checkpoint is None: # create new lora weights while training
            lora_config = LoraConfig(
                task_type=TaskType.CAUSAL_LM,
                inference_mode=False,
                r=finetuning_args.lora_rank,
                lora_alpha=finetuning_args.lora_alpha,
                lora_dropout=finetuning_args.lora_dropout,
                target_modules=finetuning_args.lora_target
            )
            model = get_peft_model(model, lora_config)

    if model_args.checkpoint_dir is not None:
        logger.info("Loaded fine-tuned model from checkpoint(s): {}".format(",".join(model_args.checkpoint_dir)))

    return model


def load_pretrained(
        model_args: ModelArguments,
        finetuning_args: FinetuningArguments,
        is_trainable: Optional[bool] = False
) -> Tuple[PreTrainedModel, PreTrainedTokenizer]:
    r"""
    Loads pretrained model and tokenizer.

    Support both training and inference.
    """
    if (not is_trainable) and model_args.checkpoint_dir is None:
        logger.warning("Checkpoint is not found at evaluation, load the original model.")
        finetuning_args = FinetuningArguments(finetuning_type="none")

    config_kwargs = {
        "trust_remote_code": True,
        "cache_dir": model_args.cache_dir,
        "revision": model_args.model_revision,
        "use_auth_token": True if model_args.use_auth_token else None,
    }

    tokenizer = AutoTokenizer.from_pretrained(
        model_args.model_name_or_path,
        use_fast=model_args.use_fast_tokenizer,
        padding_side="left",
        **config_kwargs
    )
    if tokenizer.pad_token_id is None or tokenizer.pad_token_id == 64000: # 64000 for baichuan model (older version)
        tokenizer.pad_token_id = 0 # set as the <unk> token

    config = AutoConfig.from_pretrained(model_args.model_name_or_path, **config_kwargs)
    is_mergeable = True

    # Quantization configurations (using bitsandbytes library).
    if model_args.quantization_bit is not None:
        if model_args.quantization_bit == 8:
            require_version("bitsandbytes>=0.37.0", "To fix: pip install bitsandbytes>=0.37.0")
            config_kwargs["load_in_8bit"] = True
            config_kwargs["quantization_config"] = BitsAndBytesConfig(
                load_in_8bit=True,
                llm_int8_threshold=6.0
            )
        elif model_args.quantization_bit == 4:
            require_version("bitsandbytes>=0.39.0", "To fix: pip install bitsandbytes>=0.39.0")
            require_version("transformers>=4.30.1", "To fix: pip install transformers>=4.30.1")
            require_version("accelerate>=0.20.3", "To fix: pip install accelerate>=0.20.3")
            require_version("peft>=0.4.0.dev0", "To fix: pip install git+https://github.com/huggingface/peft.git")
            config_kwargs["load_in_4bit"] = True
            config_kwargs["quantization_config"] = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_compute_dtype=model_args.compute_dtype,
                bnb_4bit_use_double_quant=model_args.double_quantization,
                bnb_4bit_quant_type=model_args.quantization_type
            )
        is_mergeable = False
        config_kwargs["device_map"] = {"": int(os.environ.get("LOCAL_RANK", "0"))}
        logger.info("Quantizing model to {} bit.".format(model_args.quantization_bit))

    if not is_trainable: # `device_map=auto` should be used for inference only
        config_kwargs["device_map"] = "auto"

    if model_args.checkpoint_dir is not None and finetuning_args.finetuning_type == "full":
        model_to_load = model_args.checkpoint_dir[0]
    else:
        model_to_load = model_args.model_name_or_path

    # Load and prepare pretrained models (without valuehead).
    model = AutoModelForCausalLM.from_pretrained(
        model_to_load,
        config=config,
        torch_dtype=torch.bfloat16 if model_args.compute_dtype == torch.bfloat16 else torch.float16,
        low_cpu_mem_usage=True,
        **config_kwargs
    )
    model = prepare_model_for_training(model, finetuning_args.finetuning_type) if is_trainable else model
    model = _init_adapter(model, model_args, finetuning_args, is_trainable, is_mergeable)

    if not is_trainable:
        model.requires_grad_(False) # fix all model params
        model = model.half() if model_args.quantization_bit is None else model # cast from fp32 to fp16

    print_trainable_params(model)

    return model, tokenizer


def plot_loss(save_dictionary: os.PathLike, keys: Optional[List[str]] = ["loss"]) -> None:
    import matplotlib.pyplot as plt
    with open(os.path.join(save_dictionary, TRAINER_STATE_NAME), "r", encoding="utf-8") as f:
        data = json.load(f)

    for key in keys:
        steps, metrics = [], []
        for i in range(len(data["log_history"])):
            if key in data["log_history"][i]:
                steps.append(data["log_history"][i]["step"])
                metrics.append(data["log_history"][i][key])

        if len(metrics) == 0:
            logger.warning(f"No metric {key} to plot.")
            continue

        plt.figure()
        plt.plot(steps, metrics, alpha=0.4, label="original")
        plt.plot(steps, smooth(metrics), label="smoothed")
        plt.title("training {} of {}".format(key, save_dictionary))
        plt.xlabel("step")
        plt.ylabel(key)
        plt.legend()
        plt.savefig(os.path.join(save_dictionary, "training_{}.png".format(key)), format="png", dpi=100)
        print("Figure saved:", os.path.join(save_dictionary, "training_{}.png".format(key)))



import json
import re

def regularize(expression):
    expression = expression.replace("m1", "-1")
    expression = expression.replace("const_", "")

    # 使用正则表达式匹配百分号数字，并将其转换为小数
    def replace_percent(match):
        percent_number = float(match.group(1))
        return str(percent_number / 100)

    # 匹配百分数
    pattern_percent = r"(\d+(\.\d*)?)%"
    result = re.sub(pattern_percent, replace_percent, expression)

    return result


def extract_single_expression(expression):
    # 将单独的表达式 op(a,b) 抽离出 op,a,b
    parts = expression.split("(")
    operator = parts[0].strip()
    arguments = parts[1].split(")")[0].split(",")
    arguments = [arg.strip() for arg in arguments]
    return operator, arguments[0], arguments[1]


def program_to_expression(program):
    if "table_" in program: # 需要通过表格推理的table_任务不处理
        return program
    program = regularize(program)
    expressions = program.split("),")
    last_res = ""
    result_history = []
    for expression in expressions:
        op, arg1, arg2 = extract_single_expression(expression)
        if arg1[0] == "#":
            arg1_num = int(arg1[1:])
            arg1 = result_history[arg1_num]
        if arg2[0] == "#":
            arg2_num = int(arg2[1:])
            arg2 = result_history[arg2_num]

        this_res = ""
        if op == "add":
            this_res = arg1 + "+" + arg2
        elif op == "subtract":
            this_res = arg1 + "-" + arg2
        elif op == "multiply":
            this_res = arg1 + "*" + arg2
        elif op == "divide":
            this_res = arg1 + "/" + arg2
        elif op == "exp":
            this_res = arg1 + "**" + arg2
        elif op == "greater":
            this_res = arg1 + ">" + arg2

        last_res = "(" + this_res + ")"
        result_history.append(last_res)
    return last_res


def process_file(data_path, out_path):
    with open(data_path, 'r') as file:
        data = json.load(file)

    with open(out_path, 'w+') as file:
        outputs = []
        for one in data:
            program = one["output"]
            expression = program_to_expression(program)
            # if "table_" not in expression:
            #     print(eval(expression))
            output = one
            output["program"] = program
            output["output"] = expression
            outputs.append(output)
        json.dump(outputs, file, indent=4)


if __name__ == '__main__':
    program1 = "subtract(6876,6876), divide(#0,6876), add(1,#1), exp(#2,const_2), multiply(#3,6876)"  # 6876.0
    print(program_to_expression(program1))
    program2 = "multiply(3176, const_m1), divide(#0, 10379)"
    print(program_to_expression(program2))
    program3 = "subtract(36.5%, 19.3%)"
    print(program_to_expression(program3))
    program4 = "subtract(78.0, 75.3), subtract(58.0, 49.9), subtract(54.0, 51.8), add(#0, #1), add(#2, #3)"
    print(program_to_expression(program4))
    program5 = "subtract(1018.5, 907.0), subtract(1033.9, 792.9), greater(#0, #1)"
    print(program_to_expression(program5))
    program6 = "table_max(cumulative foreign currency translation, none)"
    print(program_to_expression(program6))

    # data_path = "../data/train_instruction.json"
    # out_path = "../data/test_ljh_new.json"
    # process_file(data_path,out_path)


import numpy as np
import torch
import sys
import random
import datetime
import io
from dataclasses import dataclass, field


@dataclass
class ICLArguments:
    random_seed: int = field(default=42)
    resume_id: int = field(default=0)
    max_length: int = field(default=256)
    api_time_interval: float = field(default=1.0)
    temperature: float = field(default=0.1)
    logging_dir: str = field(default='./log')
    output_path: str = field(default='./data/test_predictions.json')
    use_table_markdown: bool = field(default=True)
    use_table_flatten: bool = field(default=False)
    truncation: int = field(default=3000)


class Logger(object):
    def __init__(self, logging_path, stream=io.TextIOWrapper(sys.stdout.buffer, encoding='utf8')):
        self.terminal = stream
        self.log = open(logging_path, 'a', encoding='utf-8')

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        pass


def fix_seed(seed):
    # random
    random.seed(seed)
    # Numpy
    np.random.seed(seed)
    # Pytorch
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def print_time(return_flag=0):
    t_delta = datetime.timedelta(hours=9)
    JST = datetime.timezone(t_delta, 'JST')
    now = datetime.datetime.now(JST)
    now = now.strftime('%Y/%m/%d %H:%M:%S')
    if return_flag == 0:
        print(now)
    elif return_flag == 1:
        return now.replace("/", "_").replace(" ", "").replace(":", "_")
    else:
        pass



from typing import List, Optional
from dataclasses import dataclass
from abc import abstractmethod, ABC


@dataclass
class Template(ABC):
    @abstractmethod
    def __post_init__(self) -> None:
        """
        call the self._register_template to define the template like:
        self._register_template(
            prefix="the prefix if the dataset's prefix is not defined",
            prompt="the model input with {query}",
            sep="\n",
            use_history=False
        )
        :return: NoneType
        """
        pass

    def get_prompt(self, query: str, history: Optional[list] = None, prefix: Optional[str] = "") -> str:
        r"""
        Returns a string containing prompt without response.
        """
        return "".join(self._format_example(query, history, prefix))

    def get_dialog(self, query: str, resp: str, history: Optional[list] = None, prefix: Optional[str] = "") -> List[str]:
        r"""
        Returns a list containing 2 * n elements where the 2k-th is a query and the (2k+1)-th is a response.
        """
        return self._format_example(query, history, prefix) + [resp]

    def _register_template(self, prefix: str, prompt: str, sep: str, use_history: Optional[bool] = True) -> None:
        self.prefix = prefix
        self.prompt = prompt
        self.sep = sep
        self.use_history = use_history

    def _format_example(self, query: str, history: Optional[list] = None, prefix: Optional[str] = "") -> List[str]:
        prefix = prefix if prefix else self.prefix  # use prefix if provided
        prefix = prefix + self.sep if prefix else ""  # add separator for non-empty prefix
        history = history if (history and self.use_history) else []
        history = history + [(query, "<dummy>")]
        convs = []
        for turn_idx, (user_query, bot_resp) in enumerate(history):
            if turn_idx == 0:
                convs.append(prefix + self.prompt.format(query=user_query))
                convs.append(bot_resp)
            else:
                convs.append(self.sep + self.prompt.format(query=user_query))
                convs.append(bot_resp)
        return convs[:-1]  # drop last


class defaultTemplate(Template):
    def __post_init__(self):
        """
        Default template.
        """
        self._register_template(
            prefix="A chat between a curious user and an artificial intelligence assistant. "
                   "The assistant gives helpful, detailed, and polite answers to the user's questions.",
            prompt="Human: {query}\nAssistant: ",
            sep="\n",
            use_history=False
        )



# This file is autogenerated by the command `make fix-copies`, do not edit.
from ..file_utils import requires_tf


class TensorFlowBenchmarkArguments:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TensorFlowBenchmark:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


def tf_top_k_top_p_filtering(*args, **kwargs):
    requires_tf(tf_top_k_top_p_filtering)


TF_ALBERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFAlbertForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAlbertPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_MODEL_FOR_CAUSAL_LM_MAPPING = None


TF_MODEL_FOR_MASKED_LM_MAPPING = None


TF_MODEL_FOR_MULTIPLE_CHOICE_MAPPING = None


TF_MODEL_FOR_PRETRAINING_MAPPING = None


TF_MODEL_FOR_QUESTION_ANSWERING_MAPPING = None


TF_MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING = None


TF_MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING = None


TF_MODEL_FOR_TOKEN_CLASSIFICATION_MAPPING = None


TF_MODEL_MAPPING = None


TF_MODEL_WITH_LM_HEAD_MAPPING = None


class TFAutoModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForCausalLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForSeq2SeqLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFAutoModelWithLMHead:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBartForConditionalGeneration:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBartModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_BERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFBertEmbeddings:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFBertForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertForNextSentencePrediction:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFBertForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFBertForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFBertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBertPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFBlenderbotForConditionalGeneration:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_CAMEMBERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFCamembertForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCamembertForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCamembertForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCamembertForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCamembertForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCamembertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_CTRL_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFCTRLLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCTRLModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFCTRLPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_DISTILBERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFDistilBertForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFDistilBertPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_ELECTRA_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFElectraForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFElectraForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFElectraForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFElectraForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFElectraForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFElectraForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFElectraModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFElectraPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_FLAUBERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFFlaubertForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFlaubertForQuestionAnsweringSimple:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFlaubertForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFlaubertForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFlaubertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFlaubertWithLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_FUNNEL_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFFunnelBaseModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFFunnelModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_GPT2_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFGPT2DoubleHeadsModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFGPT2LMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFGPT2MainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFGPT2Model:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFGPT2PreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_LONGFORMER_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFLongformerForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFLongformerForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFLongformerModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFLongformerSelfAttention:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


TF_LXMERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFLxmertForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFLxmertMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFLxmertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFLxmertPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFLxmertVisualFeatureEncoder:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFMarianMTModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMBartForConditionalGeneration:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_MOBILEBERT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFMobileBertForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertForNextSentencePrediction:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertForPreTraining:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFMobileBertPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_OPENAI_GPT_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFOpenAIGPTDoubleHeadsModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFOpenAIGPTLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFOpenAIGPTMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFOpenAIGPTModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFOpenAIGPTPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFPegasusForConditionalGeneration:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_ROBERTA_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFRobertaForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFRobertaPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_T5_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFT5ForConditionalGeneration:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFT5Model:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFT5PreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_TRANSFO_XL_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFAdaptiveEmbedding:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFTransfoXLLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFTransfoXLMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFTransfoXLModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFTransfoXLPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFSequenceSummary:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFSharedEmbeddings:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


def shape_list(*args, **kwargs):
    requires_tf(shape_list)


TF_XLM_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFXLMForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMForQuestionAnsweringSimple:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFXLMModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMWithLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_XLM_ROBERTA_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFXLMRobertaForMaskedLM:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMRobertaForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMRobertaForQuestionAnswering:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMRobertaForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMRobertaForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLMRobertaModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


TF_XLNET_PRETRAINED_MODEL_ARCHIVE_LIST = None


class TFXLNetForMultipleChoice:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetForQuestionAnsweringSimple:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetForSequenceClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetForTokenClassification:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetLMHeadModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetMainLayer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class TFXLNetPreTrainedModel:
    def __init__(self, *args, **kwargs):
        requires_tf(self)

    @classmethod
    def from_pretrained(self, *args, **kwargs):
        requires_tf(self)


class AdamWeightDecay:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class GradientAccumulator:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


class WarmUp:
    def __init__(self, *args, **kwargs):
        requires_tf(self)


def create_optimizer(*args, **kwargs):
    requires_tf(create_optimizer)


class TFTrainer:
    def __init__(self, *args, **kwargs):
        requires_tf(self)



import json
import gzip
import os
from PIL import Image
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm
from transformers import AutoTokenizer
import torch

ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")

MAX_LEN = 512

_CITATION = """\
@article{talmor2021multimodalqa,
  title={MultiModalQA: Complex Question Answering over Text, Tables and Images},
  author={Talmor, Alon and Yoran, Ori and Catav, Amnon and Lahav, Dan and Wang, Yizhong and Asai, Akari and Ilharco, Gabriel and Hajishirzi, Hannaneh and Berant, Jonathan},
  journal={arXiv preprint arXiv:2104.06039},
  year={2021}
}
"""

_DESCRIPTION = """\
This dataset is obtained from the official release of the MMQA.
"""

_HOMEPAGE = "https://github.com/allenai/multimodalqa"

_LICENSE = "MIT License"

_TRAINING_FILE = "MMQA_train.jsonl.gz"
_DEV_FILE = "MMQA_dev.jsonl.gz"
_TEST_FILE = "MMQA_test.jsonl.gz"
_TEXTS_FILE = "MMQA_texts.jsonl.gz"
_TABLES_FILE = "MMQA_tables.jsonl.gz"
_PASSAGE_FILE = "MMQA_texts.jsonl.gz"
_IMAGES_INFO_FILE = "MMQA_images.jsonl.gz"
_IMAGES_FILE = "final_dataset_images"
_DATA_PATH = "/home/lfy/lwh/data/mmqa"


class ClassifyDataset(Dataset):

    def __init__(self, tokenizer, type_fn , type_map, data_path = _DATA_PATH , split = 'train') -> None:
        super(ClassifyDataset,self).__init__()
        self.tokenizer = tokenizer
        self.split = split
        self.type_fn = type_fn
        self.type_map = type_map.copy()
        table_path = os.path.join(data_path, _TABLES_FILE)
        passage_path = os.path.join(data_path, _PASSAGE_FILE)
        image_path = os.path.join(data_path, _IMAGES_INFO_FILE)

        tables = {}
        with gzip.open(table_path, 'r') as f: # 表格
            for line in f:
                table = json.loads(line)
                tables[table["id"]] = table
        self.tables = tables

        texts = {}
        with gzip.open(passage_path, 'r') as f: # 文章，段落
            for line in f:
                text = json.loads(line)
                texts[text["id"]] = text
        self.texts = texts
        
        images = {}
        with gzip.open(image_path, 'r') as f: # 图片
            for line in f:
                image = json.loads(line)
                images[image["id"]] = image
        self.images = images
        if split == "train":
            file_path = os.path.join(data_path, _TRAINING_FILE)
        elif split == "dev":
            file_path = os.path.join(data_path, _DEV_FILE)
        elif split == "test":
            file_path = os.path.join(data_path, _TEST_FILE)
        else:
            raise ValueError("Invalid split name")
        
        with gzip.open(file_path, 'r') as f: # 问题
            self.data = [json.loads(line) for line in f]

        for data in tqdm(self.data):
            question_ids = self.tokenizer.encode(data['question'])
            data['input_ids'] = question_ids


    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        # TODO: add table title, image title, passage title
        data_item = self.data[index]
        meta_data = data_item['metadata']
        qid = data_item['qid']

        if self.split in ["train", "dev"]:
            label = self.type_map[self.type_fn(meta_data['type'])]
            return data_item['input_ids'], label, qid
        else:
            return data_item['input_ids'], None, qid

def collate(data, tokenizer, max_bert_len, test=False):
    bs = len(data) # batchsize
    input_ids, labels, qids = zip(*data)
    input_ids = list(input_ids)

    if not test:
        labels = list(labels)
        labels_tensor = torch.tensor(labels)
        # labels_tensor = torch.zeros(bs,len(LABELS.keys()))
        # for i in range(bs):
        #     labels_tensor[i][labels[i]] = torch.tensor(1)

    max_len = max([len(ids) for ids in input_ids])
    if max_len > max_bert_len:
        max_len = max_bert_len
    for i in range(len(input_ids)):
        input_ids[i] = input_ids[i][:max_len]
        input_ids[i] = input_ids[i] + [tokenizer.pad_token_id] * (max_len - len(input_ids[i]))
    
    input_mask = torch.where(torch.tensor(input_ids) == tokenizer.pad_token_id, torch.tensor(0), torch.tensor(1))
    input_ids = torch.tensor(input_ids)

    if not test:
        return {
            "input_ids": input_ids.cuda(),
            "input_mask": input_mask.cuda(),
            "labels": labels_tensor.cuda(),
            "qids": qids
        }
    else:
        return {
            "input_ids": input_ids.cuda(),
            "input_mask": input_mask.cuda(),
            "qids": qids
        }


import json
import gzip
import os
from PIL import Image
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm
from transformers import AutoTokenizer
import torch
from utils.image_stuff import get_caption_map

MAX_LEN = 512

_CITATION = """\
@article{talmor2021multimodalqa,
  title={MultiModalQA: Complex Question Answering over Text, Tables and Images},
  author={Talmor, Alon and Yoran, Ori and Catav, Amnon and Lahav, Dan and Wang, Yizhong and Asai, Akari and Ilharco, Gabriel and Hajishirzi, Hannaneh and Berant, Jonathan},
  journal={arXiv preprint arXiv:2104.06039},
  year={2021}
}
"""

_DESCRIPTION = """\
This dataset is obtained from the official release of the MMQA.
"""

_HOMEPAGE = "https://github.com/allenai/multimodalqa"

_LICENSE = "MIT License"

_TRAINING_FILE = "MMQA_train.jsonl.gz"
_DEV_FILE = "MMQA_dev.jsonl.gz"
_TEST_FILE = "MMQA_test.jsonl.gz"
_TEXTS_FILE = "MMQA_texts.jsonl.gz"
_TABLES_FILE = "MMQA_tables.jsonl.gz"
_PASSAGE_FILE = "MMQA_texts.jsonl.gz"
_IMAGES_INFO_FILE = "MMQA_images.jsonl.gz"
_IMAGES_FILE = "final_dataset_images"
_DATA_PATH = "/home/lfy/lwh/data/mmqa"



class RetrieverDataset(Dataset):
    def __init__(self, tokenizer, data_path = _DATA_PATH , split = 'train', caption_file_name = 'mmqa_captions_llava.json') -> None:
        super(RetrieverDataset,self).__init__()
        self.tokenizer = tokenizer
        self.split = split
        with open(os.path.join(data_path, caption_file_name), "r") as f:
            caption_map = json.load(f)
        self.caption_map = caption_map

        table_path = os.path.join(data_path, _TABLES_FILE)
        passage_path = os.path.join(data_path, _PASSAGE_FILE)
        image_path = os.path.join(data_path, _IMAGES_INFO_FILE)

        if split == "train":
            file_path = os.path.join(data_path, _TRAINING_FILE)
        elif split == "dev":
            file_path = os.path.join(data_path, _DEV_FILE)
        elif split == "test":
            file_path = os.path.join(data_path, _TEST_FILE)
        else:
            raise ValueError("Invalid split name")
        
        with gzip.open(file_path, 'r') as f: # 问题
            self.data = [json.loads(line) for line in f]

        for data in tqdm(self.data):
            question_ids = self.tokenizer.encode(data['question'])
            data['question_ids'] = question_ids

        tables = {}
        with gzip.open(table_path, 'r') as f: # 表格
            for id, line in enumerate(tqdm(f)):
                table = json.loads(line)
                tables[table["id"]] = table
        self.tables = tables

        texts = {}
        with gzip.open(passage_path, 'r') as f: # 文章，段落
            for id, line in enumerate(tqdm(f)):
                text = json.loads(line)
                # text['ids'] = tokenizer.encode(text['text'])
                texts[text["id"]] = text
        self.texts = texts
        
        images = {}
        with gzip.open(image_path, 'r') as f: # 图片
            for id, line in enumerate(tqdm(f)):
                image = json.loads(line)
                # image['ids'] = tokenizer.encode(caption_map[image['id']])
                images[image["id"]] = image
        self.images = images

    def get_caption_by_id(self, id):
        try:
            caption = self.caption_map[id]
        except:
            caption = ""
        return caption

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        # TODO: add table title, image title, passage title
        """
        Returns:
            question_ids: List[int]
            image_ids: Dict
            text_ids: Dict
            gold_image: List[str]
            gold_text: List[str]
        """
        data_item = self.data[index]
        meta_data = data_item['metadata']
        question_ids = data_item['question_ids']

        if self.split in ['train','dev']:
            gold_image = []
            gold_text = []
            for sup in data_item['supporting_context']:
                if sup['doc_part'] == 'text':
                    gold_text.append(sup['doc_id'])
                elif sup['doc_part'] == 'image':
                    gold_image.append(sup['doc_id'])
            
            gold_image = list(set(gold_image))
            gold_text = list(set(gold_text)) # unique
        
         
        # text_ids = meta_data['text_doc_ids']
        # image_ids = meta_data['image_doc_ids']

        # image_ids = dict()
        # text_ids = dict()
        # for id in meta_data['text_doc_ids']:
        #     text_ids[id] = self.tokenizer.encode(self.texts['id']['text'])

        # for id in meta_data['image_doc_ids']:
        #     image_ids[id] = self.tokenizer.encode(caption_map[id])
        
        image_docs = dict()
        text_docs = dict()
        for id in meta_data['text_doc_ids']:
            text_docs[id] = {
                'title': self.texts[id]['title'],
                'text': self.texts[id]['text']
            }
        
        for id in meta_data['image_doc_ids']:
            image_docs[id] = {
                'title': self.images[id]['title'],
                'text': self.get_caption_by_id(id)
            }
        
        if self.split in ["train", "dev"]:
            return question_ids, image_docs, text_docs, gold_image, gold_text
        else:
            return question_ids, image_docs, text_docs, None, None


def collate(data, tokenizer, max_bert_len, image_or_text='text', test=False):
    # print(json.dumps(data, indent=4))
    # exit(0)
    data = data[0]
    pad_id = tokenizer.convert_tokens_to_ids(tokenizer.pad_token)
    sep_id = tokenizer.convert_tokens_to_ids(tokenizer.sep_token)
    question_ids, images_docs, text_docs, gold_image, gold_text = data
    # print(question_ids)
    if image_or_text == 'text':
        docs = text_docs
        gold_docs = gold_text
    else :
        docs = images_docs
        gold_docs = gold_image
    
    row_num = len(docs)
    input_ids = []
    labels = torch.zeros(row_num)
    max_input_length = 0

    for doc_id,item in docs.items():
        input_data = question_ids.copy()  #  Question
        title = item['title']
        text = item['text']
        title_ids = tokenizer.convert_tokens_to_ids(tokenizer.tokenize(title)) 
        text_ids = tokenizer.convert_tokens_to_ids(tokenizer.tokenize(text))  
        input_data.extend(title_ids) # Title
        input_data.append(sep_id) # [SEP]
        input_data.extend(text_ids) # Passage/Image Text
        input_data.append(sep_id) # [SEP]
        max_input_length = max(max_input_length, len(input_data))
        input_ids.append(input_data)

    max_input_length = min(max_input_length, max_bert_len)

    for id,item in enumerate(input_ids):
        if len(item) > max_bert_len:
            item = item[:max_bert_len-1]
            item.append(sep_id)
        else:
            item = item + (max_input_length - len(item)) * [pad_id] # PAD is after SEP
        
        input_ids[id] = item

    
    input_ids = torch.tensor(input_ids)
    input_mask = torch.where(input_ids==tokenizer.pad_token_id, 0, 1)
    metadata = {
        "gold_docs": gold_docs,
        "docs": [doc_id for doc_id in docs.keys()]
    }
    # print(input_ids.shape)
    # print(input_mask.shape)

    if not test:
        num_gold = len(gold_docs)
        gold_list = [1/num_gold if doc_id in gold_docs else 0 for doc_id in docs.keys()]
        labels = torch.softmax(torch.tensor(gold_list), dim=-1, dtype=torch.float)
        # print(labels)
        return {
            "input_ids": input_ids.cuda(), "input_mask":input_mask.cuda(), "labels":labels.cuda(), \
            "metadata": metadata
        }
    else:
        return {
            "input_ids": input_ids.cuda(), "input_mask":input_mask.cuda(), "labels": None, \
            "metadata": metadata
        }


#!/usr/bin/env python3
import io

import pandas as pd
import lxml.etree
import lxml.html

from tinyhtml import *
from xlsxwriter import Workbook
from utils.excel import write_html_table_to_excel


"""
Default methods for exporting tables to various formats.
"""


def table_to_json(table, include_props=True):
    is_linked = table.is_linked
    data = []
    for row in table.cells:
        row_data = []
        for c in row:
            cell_data = c.serializable_props()
            if is_linked and '##[HERE STARTS THE HYPERLINKED PASSAGE]##' in cell_data["value"]:
                begin_idx = cell_data["value"].find('##[HERE STARTS THE HYPERLINKED PASSAGE]##')
                hyperlined_begin_idx = begin_idx + len('##[HERE STARTS THE HYPERLINKED PASSAGE]##')
                display_cell_value = cell_data["value"][:begin_idx]
                hyperlined_cell_value = cell_data["value"][hyperlined_begin_idx:]
                cell_data["value"] = display_cell_value
                cell_data["hyperlinked_passage"] = hyperlined_cell_value
            elif is_linked:
                cell_data["hyperlinked_passage"] = ""
            row_data.append(cell_data)
        data.append(row_data)
    j = {"data": data}

    if include_props:
        j["properties"] = table.props

    return j


def table_to_excel(table, include_props=True):
    file_stream = io.BytesIO()
    workbook = Workbook(file_stream, {"in_memory": True})
    worksheet = workbook.add_worksheet()
    write_html_table_to_excel(table, worksheet, workbook=workbook, write_table_props=include_props)
    workbook.close()

    return file_stream


def table_to_csv(table):
    df = table_to_df(table)

    # export headers only if they are not the default integer index
    export_headers = (type(df.columns) == pd.core.indexes.base.Index)

    table_csv = df.to_csv(index=False, header=export_headers)
    return table_csv


def table_to_df(table):
    table_el = _get_main_table_html(table, is_csv_output=True)
    table_html = table_el.render()
    df = pd.read_html(table_html)[0]
    return df


def table_to_html(table, displayed_props, include_props, html_format, merge = False):
    if html_format == "web" and table.props != {}:
        meta_el = _meta_to_html(table.props, displayed_props)
    elif html_format == "export" and include_props and table.props != {}:
        meta_el = _meta_to_simple_html(table.props)
    else:
        meta_el = h("div")(" ")
    meta_html = meta_el
    if table.html != "":
        if not merge:
            return (lxml.etree.tostring(lxml.html.fromstring(meta_html.render()), encoding="unicode", pretty_print=True),
            table.html)
        else:
            combine_html = lxml.etree.tostring(lxml.html.fromstring(meta_html.render()), encoding="unicode", pretty_print=True)+table.html
            return f'<div>{combine_html}</div>'

    table_el = _get_main_table_html(table)
    table_html = h('div')(table_el)
    
    if not merge:
        return (lxml.etree.tostring(lxml.html.fromstring(meta_html.render()), encoding="unicode", pretty_print=True),
        lxml.etree.tostring(lxml.html.fromstring(table_html.render()), encoding="unicode", pretty_print=True))
    else:
        return lxml.etree.tostring(lxml.html.fromstring(h('div')([meta_html,table_html]).render()), encoding="unicode", pretty_print=True)


def select_cells(table, highlighted_only, cell_ids):
    if cell_ids:
        return [[table.get_cell_by_id(int(idx)) for idx in cell_ids]]
    elif highlighted_only and table.has_highlights():
        return table.get_highlighted_cells()
    else:
        return table.get_cells()


def table_to_2d_str(cells, props, is_linked):

    prop_tokens = [f"{key}: {val}" for key, val in props.items()]
    prop_str = "===\n" + "\n".join(prop_tokens) + "\n===\n"

    cell_tokens = []
    for i, row in enumerate(cells):
        for j, cell in enumerate(row):
            if cell.is_dummy:
                continue
            if is_linked and '##[HERE STARTS THE HYPERLINKED PASSAGE]##' in cell.value:
                begin_idx = cell.value.find('##[HERE STARTS THE HYPERLINKED PASSAGE]##')
                display_cell_value = cell.value[:begin_idx]
            else:
                display_cell_value = cell.value
            cell_tokens.append(f"| {display_cell_value} ")
        cell_tokens.append(f"|\n")
    cell_str = "".join(cell_tokens).strip()

    return prop_str + cell_str


def table_to_linear(table_data, inclue_props):
    props = {}
    if inclue_props:
       props = table_data.props
    table = table_data.cells

    return table_to_2d_str(table, props, table_data.is_linked)



def _meta_to_html(props, displayed_props):
    meta_tbodies = []
    meta_buttons = []

    for key, value in props.items():
        meta_row_cls = "collapse show" if key in displayed_props else "collapse"
        aria_expanded = "true" if key in displayed_props else "false"

        # two wrappers around text required for collapsing
        wrapper = h("div")
        cells = [h("th")(wrapper(h("div")(key))), h("td")(wrapper(h("div")(value)))]

        meta_tbodies.append(h("tr", klass=[meta_row_cls, f"row_{key}", "collapsible"])(cells))
        meta_buttons.append(
            h(
                "button",
                type_="button",
                klass="prop-btn btn btn-fw",
                data_bs_toggle="collapse",
                data_bs_target=f".row_{key}",
                aria_expanded=aria_expanded,
                aria_controls=f"row_{key}",
                style="margin-right:8px; margin-top:5px"
            )(key)
        )


    meta_buttons_div = h("div", klass="prop-buttons")(meta_buttons)
    meta_tbody_el = h("tbody")(meta_tbodies)
    meta_table_el = h("table", klass="table table-sm caption-top meta-table")(meta_tbody_el)
    meta_el = h("div")(meta_buttons_div, meta_table_el)
    return meta_el


def _meta_to_simple_html(props):
    meta_trs = []
    for key, value in props.items():
        meta_trs.append([h("th")(key), h("td")(value)])

    meta_tbodies = [h("tr")(tds) for tds in meta_trs]
    meta_tbody_el = h("tbody")(meta_tbodies)
    meta_table_el = h("table", klass="table table-sm caption-top meta-table")(h("caption")("properties"), meta_tbody_el)
    return meta_table_el


def _get_main_table_html(table, is_csv_output = False):
    if is_csv_output:
        is_linked = False
    else:
        is_linked = table.is_linked
    trs = []
    for row in table.cells:
        tds = []
        for c in row:
            if c.is_dummy:
                continue

            eltype = "th" if c.is_header else "td"

            if is_linked and '##[HERE STARTS THE HYPERLINKED PASSAGE]##' in c.value:
                begin_idx = c.value.find('##[HERE STARTS THE HYPERLINKED PASSAGE]##')
                hyperlined_begin_idx = begin_idx+len('##[HERE STARTS THE HYPERLINKED PASSAGE]##')
                display_cell_value = c.value[:begin_idx]
                hyperlined_cell_value = c.value[hyperlined_begin_idx:]
                hyperlined_cell_value = add_dropdown_html(hyperlined_cell_value, c.idx)
                display_cell_value = [display_cell_value, hyperlined_cell_value]
            else:
                display_cell_value = c.value

            td_el = h(eltype, colspan=c.colspan, rowspan=c.rowspan, cell_idx=c.idx)(display_cell_value)

            # if c.is_highlighted:
            #     td_el.tag.attrs["class"] = "table-active"

            tds.append(td_el)
        trs.append(tds)

    tbodies = [h("tr")(tds) for tds in trs]
    tbody_el = h("tbody", id="main-table-body")(tbodies)
    table_el = h("table", klass="table table-sm no-footer table-bordered caption-top main-table",role="grid")(tbody_el)

    return table_el

def add_dropdown_html(text, idx):
    link_head = h(
        "a",
        klass="dropdown-toggle",
        id=f"messageDropdown-{idx}",
        href="#",
        data_bs_toggle="dropdown",
        aria_expanded="false"
    )(h("i", klass="mdi mdi-link mx-0")(""))
    link_body = h(
        "div",
        klass ="dropdown-menu",
        aria_labelledby=f"messageDropdown-{idx}"
    )(h("p", klass="drop-txt")(text))
    dropdowm_html = h("span")([link_head,link_body])

    return dropdowm_html










import pytest
from tatqa_utils import *


def test_extract_first_num_from_text():
    text = '2.3 million'
    assert extract_one_num_from_str(text) == 2.3
    text = '-2.3 million'
    assert extract_one_num_from_str(text) == -2.3
    text = '205 million'
    assert extract_one_num_from_str(text) == 205
    text = '-1,210 million'
    assert extract_one_num_from_str(text) == -1210


def test_to_num():
    text = '2.3 million'
    assert to_number(text) == 2300000
    text = '-2.3 thousand'
    assert to_number(text) == -2300
    text = '205 billion'
    assert to_number(text) == 205000000000
    text = '-1,210 million'
    assert to_number(text) == -1210000000



def test_ws_tokenize():
    text = '2.3   million'
    assert ws_tokenize(text) == ['2.3', 'million']
    text = '2.3 \nmillion'
    assert ws_tokenize(text) == ['2.3', 'million']
    text = '2.3\n\tmillion'
    assert ws_tokenize(text) == ['2.3', 'million']

def test_normalize_answer():
    assert normalize_answer('-134.12') == '-134.12'
    assert normalize_answer('134.12') == '134.12'
    assert normalize_answer('(134.12)') == '-134.12'
    assert normalize_answer('18.3%') == '0.183'



def test_is_num():
    assert is_number('$124')



import numpy as np
import copy 
def get_continuous_tag_slots(paragraph_token_tag_prediction):
    tag_slots = []
    span_start = False
    for i in range(0, len(paragraph_token_tag_prediction)):
        if paragraph_token_tag_prediction[i] != 0 and not span_start:
            span_start = True
            start_index = i
        if paragraph_token_tag_prediction[i] == 0 and span_start:
            span_start = False
            tag_slots.append((start_index, i))
    if span_start:
        tag_slots.append((start_index, len(paragraph_token_tag_prediction)))
    return tag_slots


def get_single_span_tokens_from_table_and_paragraph(table_cell_tag_prediction, table_cell_tag_prediction_score, paragraph_token_tag_prediction, paragraph_token_tag_prediction_score, global_tokens):
    paragraph_token_tag_prediction = paragraph_token_tag_prediction[0]
    paragraph_token_tag_prediction_score = paragraph_token_tag_prediction_score[0]
    tag_slots = get_continuous_tag_slots(paragraph_token_tag_prediction)
    best_result = float("-inf")
    best_combine = []
    for tag_slot in tag_slots:
        current_result = np.mean(paragraph_token_tag_prediction_score[tag_slot[0]:tag_slot[1]])
        if current_result > best_result:
            best_result = current_result
            best_combine = tag_slot
    table_cell_tag_prediction = table_cell_tag_prediction[0]
    table_cell_tag_prediction_score = table_cell_tag_prediction_score[0]
    tagged_cell_index = [i for i in range(len(table_cell_tag_prediction)) if table_cell_tag_prediction[i] != 0]
    
    if not tagged_cell_index and not best_combine:
        deep_cp = copy.deepcopy(table_cell_tag_prediction_score)
        aa = np.where(deep_cp==0)[0]
        deep_cp[aa] = float("-inf")
        index = np.argmax(deep_cp)
        return [str(global_tokens[index])]
    if not tagged_cell_index:
        return [" ".join(global_tokens[best_combine[0]: best_combine[1]])]
    if not best_combine:
        tagged_cell_tag_prediction_score = \
            [table_cell_tag_prediction_score[i] for i in tagged_cell_index]
        best_result_index = tagged_cell_index[int(np.argmax(tagged_cell_tag_prediction_score))]
        return [str(global_tokens[best_result_index])]
    
    tagged_cell_tag_prediction_score = \
        [table_cell_tag_prediction_score[i] for i in tagged_cell_index]
    tagged_cell_max_score = np.max(tagged_cell_tag_prediction_score)
    if tagged_cell_max_score > current_result:
        best_result_index = tagged_cell_index[int(np.argmax(tagged_cell_tag_prediction_score))]
        return [str(global_tokens[best_result_index])]
    else:
        return [" ".join(global_tokens[best_combine[0]: best_combine[1]])]

def get_single_span_tokens_from_question_table_paragraph(question_tag_prediction, question_tag_prediction_score, table_cell_tag_prediction,table_cell_tag_prediction_score,paragraph_token_tag_prediction, paragraph_token_tag_prediction_score, global_tokens):
    paragraph_token_tag_prediction = paragraph_token_tag_prediction[0]
    paragraph_token_tag_prediction_score = paragraph_token_tag_prediction_score[0]
    paragraph_tag_slots = get_continuous_tag_slots(paragraph_token_tag_prediction)
    paragraph_best_result = float("-inf")
    paragraph_best_combine = []
    for tag_slot in paragraph_tag_slots:
        current_result = np.mean(paragraph_token_tag_prediction_score[tag_slot[0]:tag_slot[1]])
        if current_result > paragraph_best_result:
            paragraph_best_result = current_result
            paragraph_best_combine = tag_slot
    
    question_tag_prediction = question_tag_prediction[0]
    question_tag_prediction_score = question_tag_prediction_score[0]
    question_tag_slots = get_continuous_tag_slots(question_tag_prediction)
    question_best_result = float("-inf")
    question_best_combine = []
    for tag_slot in question_tag_slots:
        current_result = np.mean(question_tag_prediction_score[tag_slot[0]:tag_slot[1]])
        if current_result > question_best_result:
            question_best_result = current_result
            question_best_combine = tag_slot

    table_cell_tag_prediction = table_cell_tag_prediction[0]
    table_cell_tag_prediction_score = table_cell_tag_prediction_score[0]
    tagged_cell_index = [i for i in range(len(table_cell_tag_prediction)) if table_cell_tag_prediction[i] != 0]

    if not tagged_cell_index and not question_best_combine and not paragraph_best_combine:
        deep_cp = copy.deepcopy(table_cell_tag_prediction_score)
        aa = np.where(deep_cp==0)[0]
        deep_cp[aa] = float("-inf")
        index = np.argmax(deep_cp)
        return [str(global_tokens[index])]
    if not tagged_cell_index and not paragraph_best_combine:
        return [" ".join(global_tokens[question_best_combine[0]: question_best_combine[1]])]
    if not tagged_cell_index and not question_best_combine:
        return [" ".join(global_tokens[paragraph_best_combine[0]: paragraph_best_combine[1]])]
    if not paragraph_best_combine and not question_best_combine:
        tagged_cell_tag_prediction_score = \
            [table_cell_tag_prediction_score[i] for i in tagged_cell_index]
        best_result_index = tagged_cell_index[int(np.argmax(tagged_cell_tag_prediction_score))]
        return [str(global_tokens[best_result_index])]
    
    q_score, t_score, p_score = float("-inf"), float("-inf"), float("-inf")
    if tagged_cell_index:
        tagged_cell_tag_prediction_score = \
            [table_cell_tag_prediction_score[i] for i in tagged_cell_index]
        t_score = np.max(tagged_cell_tag_prediction_score)
    if question_best_combine:
        q_score = question_best_result
    if paragraph_best_combine:
        p_score = paragraph_best_result
    index = np.argmax([q_score, t_score, p_score])
    if index == 0:
        return [" ".join(global_tokens[question_best_combine[0]: question_best_combine[1]])]
    elif index == 1:
        best_result_index = tagged_cell_index[int(np.argmax(tagged_cell_tag_prediction_score))]
        return [str(global_tokens[best_result_index])]
    elif index == 2:
        return [" ".join(global_tokens[paragraph_best_combine[0]: paragraph_best_combine[1]])]




def get_single_span_tokens_from_paragraph(paragraph_token_tag_prediction,
                                          paragraph_token_tag_prediction_score,
                                          global_tokens):
    paragraph_token_tag_prediction = paragraph_token_tag_prediction[0]
    paragraph_token_tag_prediction_score = paragraph_token_tag_prediction_score[0]
    tag_slots = get_continuous_tag_slots(paragraph_token_tag_prediction)
    best_result = float("-inf")
    best_combine = []
    for tag_slot in tag_slots:
        current_result = np.mean(paragraph_token_tag_prediction_score[tag_slot[0]:tag_slot[1]])
        if current_result > best_result:
            best_result = current_result
            best_combine = tag_slot
    if not best_combine:
        return []
    else:
        return [" ".join(global_tokens[best_combine[0]: best_combine[1]])]


def get_single_span_tokens_from_table(table_cell_tag_prediction,
                                      table_cell_tag_prediction_score,
                                      global_tokens):
    table_cell_tag_prediction = table_cell_tag_prediction[0]
    table_cell_tag_prediction_score = table_cell_tag_prediction_score[0]
    tagged_cell_index = [i for i in range(len(table_cell_tag_prediction)) if table_cell_tag_prediction[i] != 0]
    if not tagged_cell_index:
        return []
    tagged_cell_tag_prediction_score = \
        [table_cell_tag_prediction_score[i] for i in tagged_cell_index]
    best_result_index = tagged_cell_index[int(np.argmax(tagged_cell_tag_prediction_score))]
    return [str(global_tokens[best_result_index])]

def get_span_tokens_from_paragraph(paragraph_token_tag_prediction, global_tokens):
    span_tokens = []
    span_start = False
    paragraph_token_tag_prediction = paragraph_token_tag_prediction[0]

    for i in range(0, min(len(global_tokens), len(paragraph_token_tag_prediction))):
        if paragraph_token_tag_prediction[i] == 0:
            span_start = False
        if paragraph_token_tag_prediction[i] != 0:
            if not span_start:
                span_tokens.append([global_tokens[i]])
                span_start = True
            else:
                span_tokens[-1] += [global_tokens[i]]
    span_tokens = [" ".join(tokens) for tokens in span_tokens]
    return span_tokens

def get_span_tokens_from_question(question_tag_prediction, global_tokens):
    span_tokens = []
    span_start = False
    question_tag_prediction = question_tag_prediction[0]
    for i in range(0, min(len(global_tokens), len(question_tag_prediction))):
        if question_tag_prediction[i] == 0:
            span_start = False
        if question_tag_prediction[i] != 0:
            if not span_start:
                span_tokens.append([global_tokens[i]])
                span_start = True
            else:
                span_tokens[-1] += [global_tokens[i]]
    span_tokens = [" ".join(tokens) for tokens in span_tokens]
    return span_tokens


def get_span_tokens_from_table(table_cell_tag_prediction, global_tokens):
    table_cell_tag_prediction = table_cell_tag_prediction[0]
    span_tokens = []
    for i in range(0, len(table_cell_tag_prediction)):
        if table_cell_tag_prediction[i] != 0:
            span_tokens.append(str(global_tokens[i]))
    return span_tokens